<?php

namespace Intergo\RedisProxy\Proxies;

use Closure;
use Illuminate\Cache\CacheManager;
use Illuminate\Cache\Repository;
use Illuminate\Contracts\Redis\Connection;
use Intergo\RedisProxy\Proxies\Exceptions\CacheProxyConnectionDead;

class CacheProxy extends OperationsProxy
{
    /** @var Repository $defaultStore */
    protected $defaultStore;

    /** @var Connection */
    protected $readStore;

    /** @var CacheManager $cache */
    protected $cache;

    /** @var array $configuredStores */
    protected $configuredStores;


    /**
     * CacheProxy constructor used by service provider
     */
    public function __construct(CacheManager $cache, array $configuredStores)
    {
        $this->cache = $cache;
        $this->configuredStores = $configuredStores;
    }

    /**
     * Proxy call to cache store connection
     *
     * @param $name
     * @param $arguments
     * @return mixed|void
     */
    public function __call($name, $arguments)
    {
        return $this->proxy($name, $arguments);
    }

    /**
     * Proxy call on read or default connection based on called method with failover support on read
     *
     * @param string $method
     * @param array $parameters
     * @return mixed
     */
    protected function proxy(string $method, array $parameters)
    {
        if($method == "store" && count($parameters) > 0){
            if($parameters[0] == "redis"){
                return app(CacheProxy::class);
            }
            return app(CacheProxy::class.$parameters[0]);
        }elseif(is_null($this->defaultStore)){
            $this->defaultStore = $this->connect($this->cache,$this->configuredStores[self::DEFAULT_CONFIG_KEY]);
            if($this->configuredStores[self::READONLY_CONFIG_KEY] == $this->configuredStores[self::DEFAULT_CONFIG_KEY]){
                $this->readStore = $this->defaultStore;
            }else{
                $this->readStore = $this->connect($this->cache, $this->configuredStores[self::READONLY_CONFIG_KEY]);
            }
            if(is_null($this->readStore)){
                $this->readStore = $this->defaultStore;
            }
        }

        if (in_array($method, $this->getReadingOperationMethods())) {
            list($connected, $result) = $this->failover($this->readStore, $method, $parameters);
            if ($connected) {
                return $result;
            }
        }

        if(is_null($this->defaultStore)){
            throw new CacheProxyConnectionDead("Cache write connection is dead.");
        }

        return $this->defaultStore->{$method}(...$parameters);
    }

    /**
     * @return Repository
     */
    public function getDefaultStore()
    {
        return $this->defaultStore;
    }

    /**
     * @return Connection
     */
    public function getReadStore()
    {
        return $this->readStore;
    }

    /**
     * Define and return all data retrieval methods based on Illuminate\Contracts\Cache\Store
     * @return string[]
     */
    public function getReadingOperationMethods()
    {
        return ['get', 'many', 'has'];
    }

    /**
     * Since method contains both read $ write operations, get an item from the cache
     * using read connection, or execute the given Closure and store the result forever.
     * @param string $key
     * @param Closure $callback
     * @return mixed
     */
    public function rememberForever($key, Closure $callback)
    {
        //proxy for failover
        $value = $this->proxy('get', [$key]);

        /**
         * If the item exists in the cache we will just return this immediately and if not we will execute the given Closure
         * and cache the result of that for a given number of minutes, so it's available for all subsequent requests.
         */
        if (!is_null($value)) {
            return $value;
        }
        $this->defaultStore->forever($key, $value = $callback());
        return $value;
    }
}
