<?php

namespace Intergo\RedisProxy\Proxies;

use Illuminate\Redis\Connections\Connection;
use Illuminate\Redis\RedisManager;
use Exceptions\RedisProxyConnectionDead;
use Throwable;

class RedisProxy extends OperationsProxy
{
    /** @var Connection|null $defaulConnection */
    protected $defaulConnection = null;

    /** @var Connection|null */
    protected $readConnection = null;

    /** @var RedisManager $redis */
    protected $redis;

    /** @var array $configuredStores */
    protected $configuredStores;

    /**
     * RedisProxy constructor used by service provider
     */
    public function __construct(RedisManager $redis, array $configuredStores)
    {
        $this->redis = $redis;
        $this->configuredStores = $configuredStores;
    }

    /**
     * Proxy call to redis store connection
     *
     * @param $name
     * @param $arguments
     * @return mixed|void
     */
    public function __call($name, $arguments)
    {
        return $this->proxy($name, $arguments);
    }

    /**
     * Proxy call on read or default connection based on called method with failover support on read
     *
     * @param string $method
     * @param array $parameters
     * @return mixed
     */
    protected function proxy(string $method, array $parameters)
    {
        $connection = $this->defaulConnection;

        if($method == "connection" && count($parameters) > 0){
            return app(RedisProxy::class.$parameters[0]);
        }elseif(is_null($this->defaulConnection)){
            $this->defaulConnection = $this->connect($this->redis,$this->configuredStores[self::DEFAULT_CONFIG_KEY]);
            if($this->configuredStores[self::READONLY_CONFIG_KEY] == $this->configuredStores[self::DEFAULT_CONFIG_KEY]){
                $this->readConnection = $this->defaulConnection;
            }else{
                $this->readConnection = $this->connect($this->redis, $this->configuredStores[self::READONLY_CONFIG_KEY]);
            }
            if(is_null($this->readConnection)){
                $this->readConnection = $this->defaulConnection;
            }
        }

        if (in_array(strtolower($method), $this->getReadingOperationMethods())) {
            $connection = $this->readConnection;
        }elseif(is_null($this->defaulConnection)){
            throw new RedisProxyConnectionDead("Redis write connection is dead.");
        }

        list($connected, $result) = $this->failover($connection, $method, $parameters);
        if ($connected) {
            return $result;
        }

        if(is_null($this->defaulConnection)){
            throw new RedisProxyConnectionDead("Redis write connection is dead.");
        }

        return $this->defaulConnection->{$method}(...$parameters);
    }

    /**
     * @return bool
     */
    public function defaultConnectionHealth()
    {
        if (is_null(($this->defaulConnection))) {
            return false;
        }
        try {
            return (bool) $this->defaulConnection->ping();
        } catch (Throwable $exception) {
            return false;
        }
    }

    /**
     * @return bool
     */
    public function readConnectionHealth()
    {
        if (is_null(($this->readConnection))) {
            return false;
        }
        try {
            return (bool) $this->readConnection->ping();
        } catch (Throwable $exception) {
            return false;
        }
    }

    /**
     * Define and return all data retrieval methods
     * @return string[]
     */
    public function getReadingOperationMethods()
    {
        return [
            //strings
            'bitcount', 'get', 'getbit', 'getrange', 'mget', 'getmultiple',
            //keys
            'dump', 'exists', 'keys', 'getkeys', 'randomkey', 'type', 'ttl', 'pttl',
            //hashes
            'hexists', 'hget', 'hgetall', 'hkeys', 'hlen', 'hmget', 'hvals', 'hscan', 'hstrlen',
            //lists
            'lindex', 'lget', 'llen', 'lsize', 'lrange', 'lgetrange',
            //sets
            'scard', 'ssize', 'sismember', 'scontains', 'srandmember', 'ssscan',
            //sorted sets
            'zcard', 'zsize', 'zcount', 'zrange', 'zrangebyscore', 'zrevrangebyscore', 'zrangebylex', 'zrank', 'zrevrank', 'zrevrange', 'zscore', 'zscan'
        ];
    }
}
